// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


/**
   Realiza peticiones automaticas sobre el servidor de operaciones aritmeticas.
   
   Mediante un servicio de reparto por RoundRobin.

   El servidor de este cliente: server.p rserver.p 
*/

#include <iostream>

#include <string.h>
#include <unistd.h>

#include <anna/core/core.hpp>
#include <anna/app/functions.hpp>
#include <anna/comm/comm.hpp>

#include <anna/timex/Engine.hpp>
#include <anna/timex/Clock.hpp>

#include <anna/test/Response.hpp>
#include <anna/test/Request.hpp>

class Sender : public anna::timex::Clock {
public:
   Sender () : Clock ("Sender", (Millisecond)1000), 
      a_messageBySecond (0), 
      a_nquarter (0), 
      a_requests ("Request"), 
      a_errorCounter (0),
      a_txMessageCounter (0)
   {;}

   void setMessageBySecond (const int messageBySecond) throw () { a_messageBySecond = messageBySecond; }

   int getTxMessageCounter () const throw () { return a_txMessageCounter; }

private:
   int a_messageBySecond;
   int a_nquarter;
   int a_errorCounter;
   int a_txMessageCounter;
   ThreadData <test::Request> a_requests;

   bool tick () throw (RuntimeException);
};

class MyCommunicator : public Communicator {
public:
   MyCommunicator () : Communicator (), a_avgResponseTime (0), a_rxMessageCounter (0), a_responses ("Response")  {;}

private:
   using Communicator::eventBreakConnection;

   ThreadData <test::Response> a_responses;
   int a_avgResponseTime;
   int a_rxMessageCounter;

   void eventReceiveMessage (ClientSocket &, const Message&) throw (RuntimeException);
/*
   void eventBreakConnection (const ClientSocket&) throw ();

   void eventBreakConnection (Server* server) throw () {
      comm::Communicator::eventBreakConnection (server);
   }
*/
   void eventBreakConnection (const Service* service) throw ();
   
   static bool isOk (const test::Response& response) throw ();
};

class RRClient : public anna::comm::Application {
public:
   RRClient ();

   Service* getService () const throw () { return a_service; }
   const Sender* getSender () const throw () { return &a_sender; }

private:
   MyCommunicator a_communicator;
   anna::timex::Engine a_timeController;
   Sender a_sender;
   Service* a_service;

   void initialize () throw (RuntimeException);
   void run () throw (RuntimeException);
};

using namespace std;

int main (int argc, const char** argv)
{
   CommandLine& commandLine (CommandLine::instantiate ());
   RRClient app;

   srand (time (NULL));

   try {
      commandLine.initialize (argv, argc);
      commandLine.verify ();

      Logger::setLevel (Logger::Information);
      string traceFile ("client.");
      traceFile += anna::functions::asString ((int) getpid ());
      traceFile += ".trace";
      Logger::initialize ("arithmeticClient", new TraceWriter (traceFile.c_str (),4096000));

      app.start ();
   }
   catch (Exception& ex) {
      cout << ex.asString () << endl;
   }

   return 0;
}

RRClient::RRClient () :
   Application ("arithmeticClient", "Cliente de operaciones aritmeticas", "1.0"),
   a_communicator (),
   a_timeController ((Millisecond)1000, (Millisecond)250)
{
   CommandLine& commandLine (CommandLine::instantiate ());

   commandLine.add ("p", CommandLine::Argument::Mandatory, "Puertos de los servidores (separados por comas)");
   commandLine.add ("a", CommandLine::Argument::Mandatory, "Direccin IP Puerto en el que el servidor atiende respuestas.");
   commandLine.add ("n", CommandLine::Argument::Mandatory, "Numero de mensajes por segundo");
   commandLine.add ("trace", CommandLine::Argument::Optional, "Nivel de trazas (debug,warning, error,...)");
}

void RRClient::initialize ()
   throw (RuntimeException)
{
   CommandLine& cl (CommandLine::instantiate ());

   if (cl.exists ("trace"))
      Logger::setLevel (Logger::asLevel (cl.getValue ("trace")));

   a_sender.setMessageBySecond (cl.getIntegerValue ("n"));
      
   Network& network = Network::instantiate ();
   Tokenizer ports (cl.getValue ("p"), ",");
   int port;

   a_service = new comm::RoundRobinDelivery ("Service_Arithmetic", true);

   for (Tokenizer::const_iterator ii = ports.begin (), maxii = ports.end (); ii != maxii; ii ++) {
      port = atoi (Tokenizer::data (ii));
      a_service->attach (network.createServer (cl.getValue ("a"), port, true));
   }

   a_communicator.attach (a_service);      
}

void RRClient::run ()
   throw (RuntimeException)
{
   a_timeController.activate (a_sender);

   a_communicator.accept ();
}

void MyCommunicator::eventReceiveMessage (ClientSocket&, const Message& message)
   throw (RuntimeException)
{
  LOGMETHOD (TraceMethod tm ("MyCommunicator", "eventReceiveMessage", ANNA_FILE_LOCATION));

   test::Response& response = a_responses.get ();
   response.decode (message.getBody ());

   const anna::Millisecond now = anna::functions::millisecond ();
   const int delay =  now - (Millisecond) response.initTime;

   if (delay > 0 && isOk (response) == true) {
      a_rxMessageCounter ++;
      a_avgResponseTime += delay;

      LOGINFORMATION (
         string msg = anna::functions::asString (
            "%d %c %d = %d", response.x, response.op, response.y, response.result
         );
         msg += anna::functions::asText (" | Delay: ", delay);
         Logger::information (msg, ANNA_FILE_LOCATION);
      );
   }
   else {
      LOGWARNING (
         string msg = anna::functions::asString (
            "Flip: %d %c %d = %d", response.x, response.op, response.y, response.result
         );
         msg += anna::functions::asText (" | Message: ", message.getBody ());
         msg += anna::functions::asText (" | Delay: ", delay);
         Logger::warning (msg, ANNA_FILE_LOCATION);
      );
   }
}

void MyCommunicator::eventBreakConnection (const Service* service)
   throw ()
{
  LOGMETHOD (TraceMethod tm ("MyCommunicator", "eventBreakConnection", ANNA_FILE_LOCATION));

   if (a_rxMessageCounter == 0)
      return;

   comm::Communicator::eventBreakConnection (service);

   if (service->isAvailable () == true) {
      LOGNOTICE (Logger::notice (service->asString (), ANNA_FILE_LOCATION));
      return;
   }

   LOGNOTICE (
      RRClient& app = static_cast <RRClient&> (anna::app::functions::getApp ());   
      string msg ("Tiempo medio respuesta: ");
      msg += anna::functions::asString (a_avgResponseTime / a_rxMessageCounter);
      msg += " ms";
      msg += anna::functions::asText (" | Rx: ", a_rxMessageCounter);
      msg += anna::functions::asText (" | Tx: ", app.getSender ()->getTxMessageCounter ());
      Logger::notice (msg, ANNA_FILE_LOCATION);

      cout << msg << endl << endl;
   );
   requestStop ();
}

bool MyCommunicator::isOk (const test::Response& response) 
   throw ()
{
   if (response.op != '+' && response.op != '-' && response.op != '*' && response.op != '/') 
      return false;
      
   int result = 0;
   
   switch (response.op) {
      case '+':
         result = response.x + response.y;
         break;
      case '-':
         result = response.x - response.y;
         break;
      case '*':
         result = response.x * response.y;
         break;
      case '/':
         result = (response.y != 0) ? (response.x / response.y): 0;
         break;
   }
    
   return result == response.result;   
}

bool Sender::tick ()
   throw (RuntimeException)
{
  LOGMETHOD (TraceMethod tm (Logger::Local7, "Sender", "tick", ANNA_FILE_LOCATION));

   Service* service = static_cast <RRClient&> (anna::comm::functions::getApp ()).getService ();
   Communicator* communicator = anna::app::functions::component <Communicator> (ANNA_FILE_LOCATION);

   if (a_messageBySecond == 0) 
      throw RuntimeException ("Hay que indicar el numero de mensajes por segundo", ANNA_FILE_LOCATION);


   if (a_errorCounter > 100) {
      communicator->requestStop ();
      Logger::warning ("Terminado por errores continuos en la conexion", ANNA_FILE_LOCATION);
      return false;
   }

   test::Request& request = a_requests.get ();

   for (int n = 0; n < a_messageBySecond && communicator->hasRequestedStop () == false; n ++) {
      request.op = '+';
      request.x = rand () % 1000;
      request.y = rand () % 1000;
      request.initTime = anna::functions::millisecond ();

      try {
         service->send (request);
         a_txMessageCounter ++;
      }
      catch (RuntimeException& ex) {
         a_errorCounter ++;
         ex.trace ();
         break;
      }
   }
   
   return true;
}

