// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <iostream>
#include <fstream>

// STL
#include <string>
#include <map>

#include <anna/core/DataBlock.hpp>
#include <anna/core/util/Tokenizer.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceWriter.hpp>
#include <anna/core/RuntimeException.hpp>
#include <anna/io/Directory.hpp>
#include <anna/xml/xml.hpp>
#include <anna/diameter/stack/Engine.hpp>
#include <anna/diameter/codec/Engine.hpp>
#include <anna/diameter/codec/EngineManager.hpp>
#include <anna/diameter/codec/Message.hpp>
//#include <anna/diameter/codec/functions.hpp> // ApplicationId anna::diameter::codec::functions::getApplicationId(const anna::DataBlock &) throw(anna::RuntimeException);


using namespace anna;
using namespace anna::diameter;

anna::diameter::codec::Message G_codecMsg;

bool getDataBlockFromHexFile(const std::string &pathfile, anna::DataBlock &db) throw() {
  // Get hex string
  static char buffer[8192];
  std::ifstream infile(pathfile.c_str(), std::ifstream::in);

  if(infile.is_open()) {
    infile >> buffer;
    std::string hexString(buffer, strlen(buffer));
    // Allow colon separator in hex string: we have to remove them before processing with 'fromHexString':
    hexString.erase(std::remove(hexString.begin(), hexString.end(), ':'), hexString.end());
    LOGDEBUG(
      std::string msg = "Hex string (remove colons if exists): ";
      msg += hexString;
      anna::Logger::debug(msg, ANNA_FILE_LOCATION);
    );
    anna::functions::fromHexString(hexString, db);
    // Close file
    infile.close();
    return true;
  }

  return false;
}


void _exit(const std::string &message, int resultCode = 1) {
  if(resultCode)
    std::cerr << message << std::endl << std::endl;
  else
    std::cout << message << std::endl << std::endl;

  exit(resultCode);
}

// Decodes a diameter message coming from a datablock
bool decodeDataBlock(const anna::DataBlock &db/*, unsigned int & detectedApplicationId*/) throw() {
  try {
    G_codecMsg.clear();
    G_codecMsg.decode(db);
  } catch(RuntimeException &ex) {
    std::cerr << ex.asString() << std::endl << std::endl;
    //_exit(ex.asString());
    return false;
  }

  return true;
}

//-------------------------------------------------------------------
int main(int argc, char **argv) {
  std::string exec = argv[0];
  std::string execBN = exec.substr(exec.find_last_of("/") + 1);
  std::string filetrace = execBN + ".trace";
  std::cout << std::endl;

  //check command line arguments
  if(argc < 3) {
    std::string msg = "Usage: "; msg += exec;
    msg += " <stacks> <working directory> [--hex-only] [--xml-only] [--no-validation] [--ignore-flags] [--debug]\n\n";
    msg += "       stacks:                  <appid1,dictionary1#appid2,dictionary2#...#appidN,dictionaryN>\n";
    msg += "                                This is a list of #-separated stacks defined by a comma-separated pair <application-id,xml dictionary pathfile>\n";
    msg += "                                If only one stack is provided, application-id could be omitted and then, all the messages will be decoded with the\n";
    msg += "                                dictionary regardless the value of the application-id (the stack will be registered with id=0).\n";
    msg += "       Working directory:       Contains 'xml files' (ANNA-Diameter message format) and/or 'hex files' (hexadecimal content, colons allowed).\n";
    msg += "       --hex-only:              Only hex files are converted.\n";
    msg += "       --xml-only:              Only xml files are converted.\n";
    msg += "       --no-validation:         no validation is performed.\n";
    msg += "       --ignore-flags:          wrong flags regarding dictionary are ignored in xml representation.\n";
    msg += "       --debug:                 activates debug level traces (warning by default).\n";
    msg += "\n";
    msg += "       The batch process will translate '.xml' files into '.xml.as.hex', and '.hex' files into '.hex.as.xml'\n";
    msg += "       Take care not to include the dictionary files (which have '.xml' extension) in the working directory, or there will be an exception.\n";
    _exit(msg);
  }

  // Command-line parameters:
  std::string stacks = argv[1];
  std::string wkDir = argv[2];
  std::string optionals;
  int indx = 3;
  while(indx < argc) { optionals += " "; optionals += argv[indx]; indx++; }

  bool no_validation = (optionals.find("--no-validation") != std::string::npos);
  bool ignore_flags = (optionals.find("--ignore-flags") != std::string::npos);
  bool debug = (optionals.find("--debug") != std::string::npos);
  bool hexOnly = (optionals.find("--hex-only") != std::string::npos);
  bool xmlOnly = (optionals.find("--xml-only") != std::string::npos);
  if (hexOnly && xmlOnly) _exit("Cannot provide both '--hex-only' and '--xml-only' !!");
  bool processXml = hexOnly ? false:true;
  bool processHex = xmlOnly ? false:true;
  Logger::setLevel(debug ? Logger::Debug:Logger::Warning);
  Logger::initialize(execBN.c_str(), new TraceWriter(filetrace.c_str(), 2048000));
  anna::diameter::stack::Engine &stackEngine = anna::diameter::stack::Engine::instantiate();
  anna::diameter::codec::EngineManager &em = anna::diameter::codec::EngineManager::instantiate();
  anna::diameter::codec::Engine *ce;
  unsigned int appid = 0;


  // Register stacks:
  try {
    anna::Tokenizer stacksTok;
    stacksTok.apply(stacks, "#");
    anna::Tokenizer::const_iterator stacks_it, stack_it;

    for(stacks_it = stacksTok.begin(); stacks_it != stacksTok.end(); stacks_it++) {
      std::string stack = anna::Tokenizer::data(stacks_it);
      anna::Tokenizer stackTok;
      stackTok.apply(stack, ",");

      if(stackTok.size() == 1) {
        if(stacksTok.size() != 1)
          throw anna::RuntimeException("Application Id value is mandatory when more than one stack is going to be configured", ANNA_FILE_LOCATION);
        anna::diameter::stack::Dictionary *d = stackEngine.createDictionary(appid, stack); // the stack is the dictionary
        ce = new anna::diameter::codec::Engine("CodecEngineForUniqueStackId_0", d);
        em.registerCodecEngine(0, ce);
        break;
      }

      if(stackTok.size() != 2)
        throw anna::RuntimeException("Each stack must be in the form '<application-id>,<xml dictionary pathfile>'", ANNA_FILE_LOCATION);

      stack_it = stackTok.begin();
      unsigned int stackId = atoll(anna::Tokenizer::data(stack_it));
      stack_it++;
      std::string file = anna::Tokenizer::data(stack_it);
      anna::diameter::stack::Dictionary *d = stackEngine.createDictionary(stackId, file);
      std::string codecEngineName = anna::functions::asString("CodecEngineForStackId_%llu", stackId);
      ce = new anna::diameter::codec::Engine(codecEngineName.c_str(), d);
      em.registerCodecEngine(stackId, ce);
    }

    std::cout << "Stacks provided:          " << std::endl;
    std::cout << anna::functions::tab(stackEngine.asString(false /* light */));
    std::cout << std::endl;
    std::cout << "Working directory:        " << wkDir << std::endl;
    std::cout << "Validation:               " << (!no_validation ? "yes" : "no") << std::endl;
    std::cout << "Ignore Flags:             " << (ignore_flags ? "yes" : "no") << std::endl;
    std::cout << std::endl;
  } catch(anna::RuntimeException &ex) {
    _exit(ex.asString());
  }

  // Validation kindness
  for (anna::diameter::codec::appid_codec_engines_it it = em.begin(); it != em.end(); it++) {
    ce = it->second;
    ce->setFixMode(anna::diameter::codec::EngineImpl::FixMode::Never); // we will encode "as is" (because --no-validation is assumed as user desire)
    ce->setValidationDepth(anna::diameter::codec::EngineImpl::ValidationDepth::Complete); // complete validation for better reports
    if(no_validation) ce->setValidationMode(anna::diameter::codec::EngineImpl::ValidationMode::Never);
    if(ignore_flags) ce->ignoreFlagsOnValidation(true);
  }

  // Auxiliary variables:
  anna::DataBlock db_aux(true);
  anna::io::Directory directoryHex, directoryXml; // we separate (although it wouldn't be neccessary using setPattern each time), because we don't
                                                  // want to process the resulting xml files from hex conversion, only the xml files present at the
                                                  // beginning in the directory
  // Analyze FS:
  directoryHex.setPattern(".hex$");
  directoryHex.read(wkDir.c_str(), anna::io::Directory::Mode::FullPath);
  directoryXml.setPattern(".xml$");
  directoryXml.read(wkDir.c_str(), anna::io::Directory::Mode::FullPath);

  // Processing .hex files:
  bool anyHexConverted = false;
  if (processHex) {
    for (anna::io::Directory::const_iterator it = directoryHex.begin(); it != directoryHex.end(); it++) {
      const std::string& entry = anna::io::Directory::data (it);
      LOGDEBUG(anna::Logger::debug(entry + " is being converted to xml", ANNA_FILE_LOCATION));

      if(!getDataBlockFromHexFile(entry, db_aux))
        _exit("Error reading hex file provided");

      // Decode datablock:
      if (decodeDataBlock(db_aux)) {

        // Write conversion:
        std::string outputFile = entry + ".as.xml";
        std::ofstream out(outputFile.c_str(), std::ifstream::out);
        out << G_codecMsg.asXMLString();
        out.close();

        anyHexConverted = true;
      }
    }
  }

  // Processing .xml files:
  bool anyXmlConverted = false;
  if (processXml) {
    for (anna::io::Directory::const_iterator it = directoryXml.begin(); it != directoryXml.end(); it++) {
      const std::string& entry = anna::io::Directory::data (it);
      LOGDEBUG(anna::Logger::debug(entry + " is being converted to hex", ANNA_FILE_LOCATION));

      // Load file:
      G_codecMsg.loadXMLFile(entry);

      // Write conversion:
      std::string hexString = anna::functions::asHexString(G_codecMsg.code());
      std::string outputFile = entry + ".as.hex";
      std::ofstream out(outputFile.c_str(), std::ifstream::out);
      out.write(hexString.c_str(), hexString.size());
      out.close();

      anyXmlConverted = true;
    }
  }


  std::string msg = "Open '"; msg += filetrace; msg += "' in order to see process traces.\n";
  if (anyHexConverted) msg += "Open '*.hex.as.xml' files to see decoding results.\n";
  if (anyXmlConverted) msg += "Open '*.xml.as.hex' files to see encoding results.\n";
  _exit(msg, 0);
}

