// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_INetAddress_hpp
#define anna_comm_INetAddress_hpp

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <string>

#include <anna/core/RuntimeException.hpp>

namespace anna {

namespace xml {
class Node;
}

namespace comm {

class Device;

/**
   Network address abstraction.
*/
class INetAddress {
public:
  /**
     Constructor.
  */
  INetAddress() : a_device(NULL), a_port(-1)  {}

  /**
     Constructor.
     \param device Instance for device (address).
     \param port Port number.
  */
  INetAddress(const Device* device, const int port = -1)  : a_device(device),  a_port(port) {;}

  /**
     Copy constructor.
     \param other Source network address.
  */
  INetAddress(const INetAddress& other) : a_device(other.a_device) , a_port(other.a_port) {;}

  /**
     Returns the device (address) associated to this instance.
     \param exceptionWhenNull Exception is launched in case the device returned is NULL.
     \return Device (address) associated to this instance.
  */
  const Device* getDevice(const bool exceptionWhenNull = true) const throw(RuntimeException);

  /**
     Returns the port associated to this instance.
     \return Port associated to this instance.
  */
  int getPort() const throw() { return a_port; }

  /**
     Sets the address for this instance.
     \param device Address provided.
  */
  void setAddress(const Device* device) throw()  { a_device = device;  }

  /**
     Sets the port for this instance.
     \param port Port provided.
  */
  void setPort(const int port) throw() { a_port = port; }

  /**
     Copy operator.
     \param right Source address to be copied.
  */
  INetAddress& operator = (const INetAddress& right) throw() { a_device = right.a_device; a_port = right.a_port; return *this; }

  /**
    Comparison operator.
    \param right Source address to be compared.
    @return \em true when address provided is equal to this \em false in other case.
  */
  bool operator == (const INetAddress& right) const throw() { return a_device == right.a_device && a_port == right.a_port; }

  /**
     Returns the initialized state for this network address.
     @return \em true when initialized, \em false when not.
  */
  bool isNull() const throw() { return (a_device == NULL || a_port == -1);  }

  /**
     Clear the content for this instance.
  */
  void clear() throw() { a_device = NULL; a_port = -1; }

  /**
     Returns string with relevant information for this instance.
     @return string with relevant information for this instance.
  */
  std::string asString() const throw();

  /**
     Returns string with low-level format relevant information for this instance.
     @return string with low-level format relevant information for this instance.
  */
  std::string serialize() const throw();

  /**
     Returns XML document with relevant information for this instance.
     \param parent XML node from which created data will depend on.
     @return XML document with relevant information for this instance.
  */
  xml::Node* asXML(xml::Node* parent) const throw(RuntimeException);

private:
  const Device* a_device;
  int a_port;
};

}
}

#endif



