// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_timex_Engine_hpp
#define anna_timex_Engine_hpp

#include <pthread.h>

#include <vector>
#include <map>

#include <anna/app/Component.hpp>

#include <anna/timex/TimeEvent.hpp>

namespace anna {

class Thread;

namespace timex {

class TickProducer;
class TickConsumer;

namespace st {
class TickProducer;
}

namespace mt {
class TickProducer;
}

/**
   Controlador general de tiempos.

   El siguiente ejemplo muestra como obtener la instancia del gestor de tiempo asociado a nuestra aplicacion:

   \code

   #include <anna.timex.Engine.h>

   void cualquierFuncion (Timer* timer)
      throw (RuntimeException)
   {
      Engine* timeController = anna::component <Engine> (FILE_LOCATION);
      timerController->activate (timer);
   }

   Si el componente Engine no hubiera sido registrado (instanciado) en nuestra aplicacion el metodo
   template anna::component lanzara una excepcion.

   \endcode

   \see anna::Component
*/

class Engine : public app::Component {
public:
  /**
     Resolucion minima (en milisegundos) soportada por el controlador de tiempos.
  */
  static const Millisecond minResolution;

  /**
     Constructor.

     Asocia automatica el objeto controlador de tiempo a la instancia de nuestra aplicacin.

     @param maxTimeout maxima duracion de los eventos de tiempo que vamos a establecer.
     @param resolution Resolucion ofrecida para este controlador de tiempos. La duracion minima
     de un evento de tiempo de ser igual o mayor a este parametro.
  */
  Engine(const Millisecond & maxTimeout, const Millisecond & resolution);

  /**
     Destructor
  */
  virtual ~Engine();

  /**
     Devuelve la resolucion del controlador de tiempo indicada en el contructor.
     \return la resolucion del controlador de tiempo indicada en el contructor.
  */
  const Millisecond & getResolution() const throw() { return a_resolution; }

  /**
     Devuelve el periodo mximo de temporizacin indicado en el constructor.
     \return el periodo mximo de temporizacin indicado en el constructor.
  */
  const Millisecond & getMaxTimeout() const throw() { return a_maxTimeout; }

  /**
     Deja en suspenso la creacion de pulsos de reloj. Los temporizadores que estuvieran activos
     en este momento tendran una duracion indeterminada.
     Para activar el sistema de control de tiempo debe invocarse al metodo play.
     \warning Si el gestor de tiempo esta en modo pausa no se pueden activar eventos temporales.
  */
  void pause() throw(RuntimeException);

  /**
     Reactiva la creacion de pulsos de reloj despues de la invocacion al metodo pause
  */
  void play() throw(RuntimeException);

  /**
     Activa el evento de tiempo recibido como parametro. Si transcurre el tiempo indicado por
     TimeEvent::getTimeout y no hemos invocado a #cancel o a cualquier otro metodo que origine
     el fin del evento de tiempo (#stop o #kill) se invocaria al metodo TimeEvent::expire.

     @param timeEvent Evento de tiempo que vamos a activar.
  */
  void activate(TimeEvent* timeEvent) throw(RuntimeException);

  /**
     Activa el evento de tiempo recibido como parametro. Si transcurre el tiempo indicado por
     TimeEvent::getTimeout y no hemos invocado a #cancel o a cualquier otro metodo que origine
     el fin del evento de tiempo (#stop o #kill) se invocar�al metodo TimeEvent::expire.

     @param timeEvent Evento de tiempo que vamos a activar.
     \warning Si el gestor de tiempo esta en modo pausa no se pueden activar eventos temporales.
  */
  void activate(TimeEvent& timeEvent) throw(RuntimeException) { activate(&timeEvent); }

  /**
     Cancela la ejecucion del evento de tiempo recibido como parametro. Si el evento no esta
     activado simplemente se ignora.

     @param timeEvent Evento de tiempo que vamos a desactivar.
     \warning Si el gestor de tiempo esta en modo pausa no se pueden activar eventos temporales.
  */
  void cancel(TimeEvent* timeEvent) throw(RuntimeException);

  /**
     Cancela la ejecucion del evento de tiempo recibido como parametro. Si el evento no esta
     activado simplemente se ignora.

     @param timeEvent Evento de tiempo que vamos a desactivar.
  */
  void cancel(TimeEvent& timeEvent) throw(RuntimeException) { cancel(&timeEvent); }

  /**
     Obtiene el evento de tiempo asociado al identificador recibido.

     @param eventTimeId Identificador de tiempo del que queremos obtener la referencia al evento de tiempo.

     \return La referencia al evento de tiempo asociado al identificador recibido o NULL en caso de
     que exista ningn evento que coincida con el identificador.
  */
  TimeEvent* getTimeEvent(const TimeEvent::Id eventTimeId) throw();

  /**
     metodo que podemos reescribir para manejar el evento que nos informa de que el controlador de
     tiempo va a empezar a comprobar eventos de tiempo caducados. Por defecto no hace nada.
  */
  virtual void eventBeginQuantum() throw(RuntimeException) {;}

  /**
     metodo que podemos reescribir para manejar el evento que nos informa de que el controlador de
     tiempo a terminado de comprobar eventos de tiempo caducados. Por defecto no hace nada.

     @param quantumSize Nmero de eventos procesados en este quantum de tiempo.
  */
  virtual void eventEndQuantum(const int quantumSize) throw(RuntimeException) {;}

  /**
     Devuelve una cadena con la informacion mas relevante de esta instancia.
     \return Una cadena con la informacion mas relevante de esta instancia.
  */
  virtual std::string asString() const throw();

  /**
     Devuelve un documento XML con la informacion mas relevante de esta instancia.
     \param parent Nodo XML del que colgar la informacion referente a esta instancia.
     \return Un documento XML con la informacion mas relevante de esta instancia.
  */
  virtual xml::Node* asXML(xml::Node* parent) const throw();

  /**
     Devuelve la cadena por la que podemos buscar el componente.
     \return La cadena por la que podemos buscar el componente.
     \see Application::find
  */
  static const char* getClassName() { return "anna::timex::Engine"; }

private:
  typedef std::vector <TimeEvent*> Quantum;
  typedef Quantum::iterator quantum_iterator;
  typedef std::map <TimeEvent::Id, Quantum*> Directory;

  Quantum* a_timeTable;
  Engine::Directory a_directory;
  int a_currentQuantum;
  int a_maxQuantum;
  Millisecond a_resolution;
  Millisecond a_maxTimeout;
  bool a_isActive;
  TickProducer* a_tickProducer;
  TickConsumer* a_tickConsumer;
  Thread* a_thread;
  Quantum* a_expiredQuantum;
  Quantum a_delayedQuantum;
  pthread_t a_threadProducer;

  void do_initialize() throw(RuntimeException);
  void do_stop() throw();
  void kill() throw();

  void tick() throw(RuntimeException);
  int getQuantum(const int timeout) const throw() { return (a_currentQuantum + (timeout / a_resolution)) % a_maxQuantum; }
  void calibrate() throw();

  // Reimplementado de app::Component
  void do_cloneParent() throw();
  void do_cloneChild() throw(RuntimeException);

  static void notifyRelease(TimeEvent* timeEvent) throw();
  static TimeEvent* timeEvent(quantum_iterator& ii) throw() { return *ii; }

  friend class TickProducer;
  friend class TickConsumer;
  friend class st::TickProducer;
  friend class mt::TickProducer;
};

}
}

#endif


