// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/core/tracing/Logger.hpp>
#include <anna/config/defines.hpp>
#include <anna/core/functions.hpp>

#include <anna/http/Transport.hpp>
#include <anna/http/internal/Tokenizer.hpp>
#include <anna/http/internal/Token.hpp>
#include <anna/http/internal/sccs.hpp>
#include <anna/http/internal/defines.hpp>
#include <anna/http/MessageFactory.hpp>
#include <anna/http/Method.hpp>
#include <anna/http/internal/EncodedBlock.hpp>

#include <anna/http/parser/Abstract.hpp>

using namespace std;
using namespace anna;

comm::TransportFactoryImpl <http::Transport> http::Transport::st_factory;

http::Transport::Transport() :
  comm::Transport(),
  a_parser(NULL),
  a_inputMessage(NULL),
  a_result(false),
  a_haveToClear(false),
  a_encodedBlock(NULL),
  a_lastChunkedByte(0),
  a_fullMessage(NULL) {
  http::sccs::activate();
  clear();
  setOverQuotaSize(8192);
  activateTimeout();
}

//----------------------------------------------------------------------------------------------
// Inicializa el estado de esta instancia. Libera el ultimo mensaje HTTP recibido,
// inicializa el estado del Parser y las variables numericas que indican atributos
// importantes del mensaje HTTP recibido.
//----------------------------------------------------------------------------------------------
void http::Transport::clear()
throw() {
  comm::Transport::clear();

  if(a_inputMessage != NULL) {
    MessageFactory::instantiate().release(a_inputMessage);
    a_inputMessage = NULL;
  }

  parser::Abstract::setState(*this, parser::Abstract::ClassType::WaitMessage);
  a_contentLength = a_bodyOffset = -1;
  a_lastChunkedByte = 0;
  a_haveToClear = false;

  if(a_encodedBlock != NULL)
    a_encodedBlock->clear();

  a_fullMessage = NULL;
}

http::Message* http::Transport::getInputMessage()
throw(RuntimeException) {
  if(a_inputMessage == NULL)
    throw RuntimeException("getInputMessage | HTTP message was not extracted", ANNA_FILE_LOCATION);

  return a_inputMessage;
}

void http::Transport::setParserState(const parser::Abstract* parser)
throw(RuntimeException) {
  if(a_parser == parser)
    return;

  a_parser = parser;
  LOGDEBUG(
    string msg("http::Transport::setParserState | Transport: ");
    msg += functions::asHexString(anna_ptrnumber_cast(this));
    msg += " | ";
    msg += a_parser->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

http::Message* http::Transport::allocateInputMessage(const http::Message::Type::_v type)
throw(RuntimeException) {
  if(a_inputMessage != NULL)
    throw RuntimeException("http::Transport::allocateInputMessage | Former HTTP message was not released", ANNA_FILE_LOCATION);

  return a_inputMessage = MessageFactory::instantiate().create(type);
}

const http::Tokenizer& http::Transport::split(const http::Token& token)
throw(RuntimeException) {
  a_lineScope.apply(token);
  return a_lineScope;
}

const http::Tokenizer& http::Transport::split(const http::Token& token, const char* separator)
throw(RuntimeException) {
  a_lineScope.apply(token, separator);
  return a_lineScope;
}

const http::Tokenizer& http::Transport::split(const http::Token& token, const char separator)
throw(RuntimeException) {
  a_lineScope.apply(token, separator);
  return a_lineScope;
}

/* El primer bloque se crea con el tamao aportado por el exterior
 * El segundo bloque se crea con el tamao calculado al analizar el mensaje
 */
const http::Message* http::Transport::externalDecode(const char* buffer, const int size)
throw(RuntimeException) {
  DataBlock dataBlock(buffer, size, false);
  clear();
  int httpSize;

  if((httpSize = calculeSize(dataBlock)) == -1) {
    DataBlock shortDataBlock(buffer, min(128, size), false);
    string msg("http::Transport::externalDecode | Buffer: ");
    msg += anna::functions::asString(shortDataBlock);
    msg += " | Can not calculate length of message";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  if(httpSize > size) {
    DataBlock shortDataBlock(buffer, min(128, size), false);
    string msg("http::Transport::externalDecode | ");
    msg += functions::asString("Message is incomplete | Size: %d | HTTP-Size: %d | Buffer:", size, httpSize);
    msg += anna::functions::asString(shortDataBlock);
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  DataBlock fine(buffer, httpSize, false);
  return static_cast <const http::Message*>(decode(fine));
}

//----------------------------------------------------------------------------------------------
// Segn vayan pasando por los estados se iran estableciendo la longitud del mensaje, el punto
// de inicio, etc ...
//
// Este metodo se puede invocar N-veces para cada mensaje.
//
// Analiza independientemente cada linea.
//
// (1) Si comenzamos un nuevo mensaje tenemos que iniciar el estado de la instancia.
// (2) Si estamos analizando un Transfer-Encoding: chunked.
//----------------------------------------------------------------------------------------------
int http::Transport::calculeSize(const DataBlock& dataBlock)
throw(RuntimeException) {
  Tokenizer::iterator ii;
  Tokenizer::iterator maxii;
  int result = -1;

  if(a_haveToClear == true)                                                    // (1)
    clear();

  // Hay estados del parser que requieren acceso al mensaje completo.
  a_fullMessage = &dataBlock;

  /* Recordar que el 'dataBlock' contiene TODO el mensaje que llevamos acumulando y eso
   * es malo para analizar los distintos chunk's que nos van llegando.
   * As que lo que hace en caso de Transfer-Encoding: chunked es "saltar" todos los bytes
   * que hayan sido procesados en alguna de las llamadas anteriores.
   */
  if(a_encodedBlock != NULL && a_encodedBlock->isValid() && a_lastChunkedByte > 0) {                  // (2)
    const int chunkSize = dataBlock.getSize() - a_lastChunkedByte;
    DataBlock chunk(dataBlock.getData() + a_lastChunkedByte, chunkSize, false);
    a_fullScope.apply(chunk, endOfLine);
    LOGDEBUG(
      string msg("http::Transport::calculeSize | DataBlockSize: ");
      msg += anna::functions::asString("%d | LastChunkSize: %d | Chunk: ", dataBlock.getSize(), a_lastChunkedByte);
      msg += functions::asString(chunk, 24);
      Logger::debug(msg, ANNA_FILE_LOCATION);
    );
    a_fullScope.apply(chunk, endOfLine);

    for(ii = a_fullScope.begin(), maxii = a_fullScope.end(); ii != maxii && result == -1; ii ++) {
      Token& token = *Tokenizer::token(ii);
      result = a_parser->processLine(*this, chunk, token);
    }
  } else {
    a_fullScope.apply(dataBlock, endOfLine);

    for(ii = a_fullScope.begin(), maxii = a_fullScope.end(); ii != maxii && result == -1; ii ++) {
      Token& token = *Tokenizer::token(ii);
      LOGDEBUG(
        string msg("http::Transport::calculeSize | Line: ");
        msg += functions::asString(token, 24);
        Logger::debug(msg, ANNA_FILE_LOCATION);
      )
      result = a_parser->processLine(*this, dataBlock, token);
    }
  }

  return result;
}

//----------------------------------------------------------------------------------------------
// Una vez que tenemos todo el mensaje HTTP completo, solo queda establecer el cuerpo del
// mensaje, si es que existe.
//
// (0) El estado parser::WaitMessage deberia haber instancia algun tipo de mensaje.
// (1) Hemos terminado de procesar un mensaje => la proxima vez que llamemos a calculeSize sera
// para comenzar con el analisis de un un nuevo mensaje.
//----------------------------------------------------------------------------------------------
const comm::Message* http::Transport::decode(const DataBlock& dataBlock)
throw(RuntimeException) {
  const bool isEncoded = (a_encodedBlock == NULL) ? false : a_encodedBlock->isValid();

  if(isEncoded == true) {
    switch(a_encodedBlock->getType()) {
    case EncodedBlock::Type::Chunked:
      a_inputMessage->comm::Message::setBody(*a_encodedBlock);
      break;
    default:
      a_inputMessage->clearBody();
      break;
    }
  } else {
    int maxSize = dataBlock.getSize();

    if(a_bodyOffset >= maxSize) {
      string msg("anna::http::Transport::decode | Error detected | BodyOffset: ");
      msg += functions::asString(a_bodyOffset);
      msg += " | Data: ";
      msg += functions::asString(dataBlock);
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    if(a_inputMessage == NULL) {                                                  // (0)
      string msg("anna::http::Transport::decode | HTTP message was not extracted | Data: ");
      msg += functions::asString(dataBlock);
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    if(a_bodyOffset == -1)
      a_inputMessage->clearBody();
    else {
      int bodySize = maxSize - a_bodyOffset;

      if(bodySize > 0)
        a_inputMessage->comm::Message::setBody(dataBlock.getData() + a_bodyOffset, bodySize);
      else
        a_inputMessage->clearBody();
    }
  }

  a_haveToClear = true;                                                         // (1)
  return a_inputMessage;
}

//-----------------------------------------------------------------------------------------------------
// Toda la complejidad de la codificacion la soporta el http::Message.
//-----------------------------------------------------------------------------------------------------
const DataBlock& http::Transport::code(comm::Message& message)
throw(RuntimeException) {
  return message.code();
}

http::EncodedBlock* http::Transport::getEncodedBlock()
throw() {
  if(a_encodedBlock == NULL)
    a_encodedBlock = new EncodedBlock;

  return a_encodedBlock;
}
